;;; smartcap.el --- try to capitalize titles cleverly  -*- lexical-binding: t; -*-

;; Copyright (C) 2015  Rasmus Pank Roulund

;; Author: Rasmus Pank Roulund <emacs@pank.eu>
;; Keywords: convenience, files

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; Try to capitalize word smartly.  Currently supports English org and
;; message mode.

;;; TODOs:

;; Probably doesn't work with e.g. subword-moe.

;;; Code:

(defcustom smartcap-uncapitalized '(("en" . ("a" "ago" "an" "and" "as" "at" "but" "by" "for"
                                             "from" "in" "into" "it" "next" "nor" "of" "off"
                                             "on" "onto" "or" "over" "past" "so" "the" "till"
                                             "to" "up" "yet")))
  "Alist of uncapitalized language.
Each cell is of the form (LANGUAGE . LIST-OF-UNCAPITALIZED-WORDS).")

(defcustom smartcap-mode-title-line-identifiers
  '((org-mode . smartcap-org-title-line)
    (message-mode . smartcap-message-title-line))
  "Alist of major-mode and their title-line identifiers.")

(defcustom smartcap-language-identifier 'smartcap-ispell-language
  "Function to find current language.")

(defcustom smartcap-language-default "en"
  "Default language")

(defun smartcap-ispell-language ()
  (let ((lang (or ispell-local-dictionary ispell-dictionary)))
    (and lang (downcase lang))))

(defun smartcap--get-uncapitalized (language)
  (cdr (assoc (cl-case (intern language)
                ((en en_us en_gb) "en")
                (t nil))
              smartcap-uncapitalized)))

(defun smartcap-org-title-line ()
  "Determine if line is a title line in `org-mode'."
  (save-excursion
    (beginning-of-line)
    (if  (looking-at (format "\\(?:%s\\)"
                             (mapconcat 'identity 
                                        `(,org-complex-heading-regexp
                                          "^\\(?10:#\\+\\(?:subtitle\\|title\\):\\)") "\\|")))
        `((title-begin . ,(match-end 0))
          (first-word . ,(progn (goto-char (or (match-beginning 4) (match-end 10)))
                                (skip-chars-forward " \t")
                                (cons (point) (progn (forward-word) (point)))))
          (last-word . ,(progn (goto-char (or (match-end 4)
                                              (line-end-position)))
                               (skip-chars-backward " \t")
                               (cons (save-excursion (forward-word -1) (point))
                                     (point))))))))

(defun smartcap-message-title-line ()
  "Determine if line is a title line in `message-mode'"
  (save-excursion 
    (when (< (point)
             (save-excursion (message-goto-body)
                             (line-beginning-position)))
      (beginning-of-line)
      (if (looking-at "^Subject:")
          `((title-begin . (match-end 0))
            (first-word  . ,(progn (goto-char (match-end 0))
                                   (skip-chars-forward " \t")
                                   (cons (point) (progn (forward-word) (point)))))
            (last-word . ,(progn (end-of-line)
                                 (skip-chars-backward " \t")
                                 (cons (save-excursion (forward-word -1) (point))
                                       (point)))))))))

(defun turn-smartcap-on ()
  ;; TODO doesn't work with e.g. `subword-capitalize', I guess.
  (local-set-key [remap capitalize-word]
                 'smartcap-capitalize-word))

(defun smartcap-capitalize-word (&optional arg)
  "Proper English title capitalization of a marked region"
  ;; TODO: I think there's a lot of assumptions on the syntax table.
  (interactive)
  (let* ((language (funcall smartcap-language-identifier))
         (uncapitalized (smartcap--get-uncapitalized language))
         (title (and uncapitalized
                     (funcall (cdr (assoc major-mode smartcap-mode-title-line-identifiers))))))
    (if title
        (let ((first-word (cdr (assoc 'first-word title)))
              (last-word (cdr (assoc 'last-word title))))
          (cond 
           ;; Before first word.
           ((<= (point) (car first-word))
            (if (<= (cdr first-word)
                    (save-excursion (forward-word)
                                    (point)))
                (capitalize-word 1)
              (forward-word)))
           ;; Last word.
           ((and (<= (point) (car last-word))
                 (>= (save-excursion (forward-word)
                                     (point))
                     (cdr last-word)))
            (capitalize-word 1))
           ;; Capitalize or lower word in between
           ((and (>= (point) (cdr first-word))
                 (< (point) (cdr last-word)))
            (if (member-ignore-case
                 (save-excursion
                   (skip-chars-forward " \t\\s.\\s_\\s-")
                   (looking-at "\\w+")
                   (match-string 0))
                 uncapitalized)
                (downcase-word 1)
              (capitalize-word 1)))
           (t (capitalize-word 1))))
      (capitalize-word 1))))

(defun smartcap-capitalize-region (&optional begin end)
  "Capitalize region using `smartcap-capitalize-word'."
  (interactive)
  (let ((beg (or begin (region-beginning)))
        (end (or end (region-end))))
    (save-excursion
      (goto-char beg)
      (while (< (point)  end)
        (smartcap-capitalize-word)))))

(provide 'smartcap)
;;; smartcap.el ends here
